import {
  takeEvery,
  take,
  call,
  select,
  all,
  put,
} from 'redux-saga/effects';
import fetch from 'node-fetch';

import check from '../check';
import root from '../root';

import * as serverlist from '../serverlist';
import * as mastersync from '../mastersync';

let myRedux;

export const states = {
  IDLE: 0,
  LOADING: 1,
};

export const initialState = {
  state: states.IDLE,
};

function* inspectGameServer(addr) {
  check.assert.string(addr);

  const addrSplit = addr.split(':');
  const ip = addrSplit[0];
  check.assert(check.ipv4(ip));

  const port = parseInt(addrSplit[1], 10);
  check.assert(check.port(port));

  try {
    const response = yield call(fetch, `http://${ip}:${port}`);
    if (response.status !== 200) {
      throw new Error(`Unexpected HTTP error code ${response.status}: ${response.statusText}`);
    }
    const serverInfo = yield call(response.json.bind(response));
    check.assert.string(serverInfo.name);
    check.assert(check.port(serverInfo.port));
    check.assert.maybe.string(serverInfo.hostPlayer);
    check.assert.maybe.string(serverInfo.map);
    check.assert.maybe.string(serverInfo.mapFile);
    check.assert.maybe.string(serverInfo.variant);
    check.assert.maybe.string(serverInfo.variantType);
    check.assert.integer(serverInfo.maxPlayers);
    check.assert.integer(serverInfo.numPlayers);
    check.assert.inRange(serverInfo.numPlayers, 0, serverInfo.maxPlayers);
    check.assert.array(serverInfo.players);
    yield put(serverlist.actions.add({
      ip,
      port,
      ...serverInfo,
    }));
  } catch (e) {
    console.error(`Request to game server ${addr} failed:`, e);
  }
}

function* updateGameServersFromMaster(masterListUrl) {
  check.assert.string(masterListUrl);

  try {
    const response = yield call(fetch, masterListUrl);
    if (response.status !== 200) {
      throw new Error(`Server returned HTTP error code ${response.status}: ${response.statusText}`);
    }
    const {
      listVersion,
      result: {
        code,
        msg,
        servers,
      },
    } = yield call(response.json.bind(response));
    switch (listVersion) {
      case 1:
        if (code !== 0) {
          throw new Error(`Server returned error code ${code}: ${msg}`);
        }
        yield all(servers.map(addr => call(inspectGameServer, addr)));
        break;
      default:
        throw new Error(`Unsupported list version: ${listVersion}`);
    }
  } catch (e) {
    console.error(`Update from ${masterListUrl} failed:`, e);
  }
}

function* updateGameServers() {
  yield put(mastersync.actions.updateDewritoJsonStarted());

  // First let's ensure we have master servers at all
  if (!(yield select(mastersync.selectors.hasMasters))) {
    // Fetch master servers and hope for success
    yield put(mastersync.actions.updateDewritoJson());
    const { err } = yield take(mastersync.actionTypes.updateDewritoJsonFinished);
    if (err) {
      yield put(mastersync.actions.updateDewritoJsonFinished({ err }));
      return;
    }
  }

  // Now we have master servers, let's check all game servers!
  yield put(serverlist.actions.reset());
  const masterListUrls = (yield select(mastersync.selectors.getMasters))
    .map(server => server.list);
  yield all(masterListUrls
    .map(url => call(updateGameServersFromMaster, url)));
  yield put(mastersync.actions.updateDewritoJsonFinished({}));
}

myRedux = root.sub('serverlist_loader').all({
  initialState,

  actions: {
    updateGameServers: {},

    updateGameServersStarted: {
      reducer: state => ({
        ...state,
        state: states.LOADING,
      }),
    },

    updateGameServersFinished: {
      reducer: state => ({
        ...state,
        state: states.IDLE,
      }),
    },
  },

  // selectors are all function(state, ...parameters)
  selectors: {
    getState: ({ state }) => state,
  },
});

export const sagas = [
  takeEvery(myRedux.actionTypes.updateGameServers, updateGameServers),
];

export const {
  actions,
  actionTypes,
  reducer,
  selectors,
} = myRedux;

